/**
 *  =============================================================================
 *  DemoChicagoEnergyBenchmarking.java: Experimental code for our "digital twins"
 *  paper published in Engineering Management, ASCE, 2020.
 * 
 *  Written by: Mark Austin                                          Summer, 2019 
 *  =============================================================================
 */

package arraylist06;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.CSVWriter;

import java.lang.Math;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.*;

public class DemoChicagoEnergyBenchmarking {
   private static final String ADDRESS_FILE="data/ChicagoEnergyBenchmarking.csv";
   public static boolean DEBUG = true;

   // Arraylist for data items ...

   private static ArrayList<EnergyDataItem> data;

   // Arraylists for property-specific data items ...

   private static ArrayList<EnergyDataItem>       office;
   private static ArrayList<EnergyDataItem>     hospital;
   private static ArrayList<EnergyDataItem> mutltifamily;
   private static ArrayList<EnergyDataItem>      housing;
   private static ArrayList<EnergyDataItem>   university;

   // Arraylists for composite (multiple property types) list of data items ...

   private static ArrayList<EnergyDataItem> composite;

   // Map "community area" name to sets of data items ...

   private static TreeMap<String, HashSet<EnergyDataItem>> communityAreaDataItemMap;

   // Map "property type" to sets of data items ...

   private static TreeMap<String, HashSet<EnergyDataItem>> propertyTypeDataItemMap;

   // Sort arraylist members by source EUI ....
   // -----------------------------------------------------------

   static class euiCompare implements Comparator<EnergyDataItem> {
      public int compare( EnergyDataItem o1, EnergyDataItem o2) {
          EnergyDataItem p1 = (EnergyDataItem) o1;
          EnergyDataItem p2 = (EnergyDataItem) o2;

          String s1 = p1.get("Source EUI (kBtu/sq ft)");
          String s2 = p2.get("Source EUI (kBtu/sq ft)");

          double eu01 = Double.parseDouble( s1 );
          double eu02 = Double.parseDouble( s2 );

          if ( eu01 == eu02 )
               return  0;
          else if ( eu01 > eu02 )
               return  1;
          else
               return -1;
      }
   }

   // Sort arraylist members by age ....
   // -----------------------------------------------------------

   static class ageCompare implements Comparator<EnergyDataItem> {
      public int compare( EnergyDataItem o1, EnergyDataItem o2) {
          EnergyDataItem p1 = (EnergyDataItem) o1;
          EnergyDataItem p2 = (EnergyDataItem) o2;

          String s1 = p1.get("Age");
          String s2 = p2.get("Age");

          double eu01 = Double.parseDouble( s1 );
          double eu02 = Double.parseDouble( s2 );

          if ( eu01 == eu02 )
               return  0;
          else if ( eu01 > eu02 )
               return  1;
          else
               return -1;
      }
   }

   // Sort arraylist members by zip code ....
   // -----------------------------------------------------------

   static class zipCompare implements Comparator<EnergyDataItem> {
      public int compare( EnergyDataItem o1, EnergyDataItem o2) {
          EnergyDataItem p1 = (EnergyDataItem) o1;
          EnergyDataItem p2 = (EnergyDataItem) o2;

          String s1 = p1.get("ZIP Code");
          String s2 = p2.get("ZIP Code");

          double eu01 = Double.parseDouble( s1 );
          double eu02 = Double.parseDouble( s2 );

          if ( eu01 == eu02 )
               return  0;
          else if ( eu01 > eu02 )
               return  1;
          else
               return -1;
      }
   }

   // =========================================================
   // Main method ...
   // =========================================================

   public static void main(String[] args) throws IOException {

      System.out.printf("Enter DemoChicagoEnergyBenchmarking.main() .... \n" );
      System.out.printf("=============================================== \n" );

      // Create csv file reader ...

      System.out.printf("--- Read: %s ...\n", ADDRESS_FILE );

      CSVReader reader = new CSVReader(new FileReader(ADDRESS_FILE));

      // Create arraylist for energy data item ...

      data = new ArrayList<EnergyDataItem>();

      // Read column headers off first column ...

      System.out.printf("--- Read column headers off first column ...\n" );

      String [] headerLine;
      headerLine = reader.readNext();

      System.out.printf("\n");
      System.out.printf("========================================= \n" );
      System.out.printf("PART 01: Column Headers ...\n" );
      System.out.printf("========================================= \n" );

      for (int i = 0; i < headerLine.length; i = i + 1) {
         System.out.printf("Column %d : %s \n", i + 1, headerLine[i] );
      }

      // Read csv data entries ...

      System.out.printf("--- Loop over csv data entries ...\n" );

      System.out.printf("\n");
      System.out.printf("========================================= \n" );
      System.out.printf("PART 02: Read and Print CSV entries ... \n");
      System.out.printf("========================================= \n" );

      String [] nextLine;
      int itemNo = 0;
      double totalElectricityUsage = 0.0;
      double totalGasUsage  = 0.0;
      double totalSourceEUI = 0.0;

      double minValueEUI = 0.0;
      double maxValueEUI = 0.0;
      int noValuesEUI = 0;

      while ((nextLine = reader.readNext()) != null) {

          // Print details of data item ...

          itemNo = itemNo + 1;
          System.out.printf("\n");
          System.out.printf("--- Data Item %d ...\n", itemNo );

          for (int i = 0; i < headerLine.length; i = i + 1) {
             System.out.printf("---   %s: %s ...\n", headerLine [i], nextLine[i] );
          }

          // Assemble energy data item object ...

          EnergyDataItem item = new EnergyDataItem();
          for (int i = 0; i < headerLine.length; i = i + 1) {
              item.add ( headerLine [i], nextLine[i] );
          }

          // ---------------------------------------------------------------
          // Process items and remove outliers from data set. Only accept:
          // 
          // 1. 0 < Source EUI < 1,000 ...
          // 2. age < 200 ...
          // 3. year = 2016
          // ---------------------------------------------------------------

          String sourceEUI = item.get("Source EUI (kBtu/sq ft)");
          String   itemAge = item.get("Age");
          String  itemYear = item.get("Data Year");

          double age = 1000.0;
          if (itemAge.length() > 0) {
             age = Double.parseDouble( itemAge );
          }

          System.out.printf("--- ItemYear: %s \n", itemYear );

          int year = 0;
          if (itemYear.length() > 0) {
             year = Integer.parseInt( itemYear );
          }

          if (sourceEUI.length() > 0 && age <= 200.0 && year == 2016 ) {
              double sourceEUIvalue = Double.parseDouble( sourceEUI );

              if ( sourceEUIvalue > 0 && sourceEUIvalue < 1000.0 ) {

                 // Compute EUI statistics ...

                 totalSourceEUI = totalSourceEUI + sourceEUIvalue;
                 noValuesEUI    = noValuesEUI + 1;

                 if ( minValueEUI == 0 && maxValueEUI == 0 ) {
                    minValueEUI = maxValueEUI = Double.parseDouble( sourceEUI );
                 } else {
                    minValueEUI = Math.min ( minValueEUI, Double.parseDouble( sourceEUI ));
                    maxValueEUI = Math.max ( maxValueEUI, Double.parseDouble( sourceEUI ));
                 }

                 // Retrieve electricity usage from data item ...

                 String eUsage = item.get("Electricity Use (kBtu)");
                 if (eUsage.length() > 0)
                    totalElectricityUsage = totalElectricityUsage + Double.parseDouble( eUsage );

                 // Retrieve natural gas usage from data item ...

                 String gUsage = item.get("Natural Gas Use (kBtu)");
                 if (gUsage.length() > 0)
                    totalGasUsage = totalGasUsage + Double.parseDouble( gUsage );

                 // Add data item to array list ...

                 data.add (item);
             }
         }
      }

      System.out.printf("--- Print summary of energy usage ...\n" );

      double averageSourceEUI = totalSourceEUI/noValuesEUI;

      System.out.printf("\n");
      System.out.printf("Summary ...\n");
      System.out.printf("=============================================================== \n");
      System.out.printf("--- No energy data items    = %10d ...\n", data.size() );
      System.out.printf("--- Total electricity usage = %10.2e (kBtu) ...\n", totalElectricityUsage );
      System.out.printf("--- Total natural gas usage = %10.2e (kBtu) ...\n", totalGasUsage );
      System.out.printf("--- No Site EUI values      = %10d ...\n", noValuesEUI );
      System.out.printf("--- Average Source EUI      = %10.2e (kBtu/sq ft) ...\n", averageSourceEUI );
      System.out.printf("--- Minimum Source EUI      = %10.2e (kBtu/sq ft) ...\n", minValueEUI );
      System.out.printf("--- Maximum Source EUI      = %10.2e (kBtu/sq ft) ...\n", maxValueEUI );
      System.out.printf("=============================================================== \n");

      System.out.printf("\n");
      System.out.printf("===============================================================\n");
      System.out.printf("PART 03: Assemble community area/set of data points mapping ...\n");
      System.out.printf("===============================================================\n");
      System.out.printf("\n");

      communityAreaDataItemMap = new TreeMap <String, HashSet<EnergyDataItem>> ();

      for (EnergyDataItem dataitem : data) {
          String community = dataitem.get("Community Area");

          HashSet area = communityAreaDataItemMap.get( community.toUpperCase() );
          if(area == null) {
             HashSet<EnergyDataItem> hh = new HashSet<EnergyDataItem> ();
             hh.add( dataitem );
             communityAreaDataItemMap.put( community.toUpperCase(), hh);
          } else {
             area.add( dataitem );
          }
      }

      System.out.printf("\n");
      System.out.printf("==============================================================\n");
      System.out.printf("PART 04: Assemble property type/set of data points mapping ...\n");
      System.out.printf("==============================================================\n");
      System.out.printf("\n");

      propertyTypeDataItemMap = new TreeMap <String, HashSet<EnergyDataItem>> ();

      for (EnergyDataItem dataitem : data) {
          String propertyType = dataitem.get("Primary Property Type");

          HashSet area = propertyTypeDataItemMap.get( propertyType.toUpperCase() );
          if(area == null) {
             HashSet<EnergyDataItem> hh = new HashSet<EnergyDataItem> ();
             hh.add( dataitem );
             propertyTypeDataItemMap.put( propertyType.toUpperCase(), hh);
          } else {
             area.add( dataitem );
          }
      }

      System.out.printf("\n");
      System.out.printf("================================================\n");
      System.out.printf("PART 05: Data points in each community area  ...\n");
      System.out.printf("================================================\n");
      System.out.printf("\n");

      // Print summary of data items belonging to community areas ...

      Iterator<Map.Entry<String, HashSet<EnergyDataItem>>> iterator = communityAreaDataItemMap.entrySet().iterator();
      while(iterator.hasNext()) {
         Map.Entry<String, HashSet<EnergyDataItem>> entry = iterator.next();
         String key = (String) entry.getKey();
         Set<EnergyDataItem> value = (Set) entry.getValue();

         System.out.printf("--- %s has %d data points ...\n", key, value.size() );
      }

      System.out.printf("\n");
      System.out.printf("=================================================\n");
      System.out.printf("PART 06: Property type statistics ...\n");
      System.out.printf("=================================================\n");
      System.out.printf("\n");

      // Print summary of data items belonging to community areas ...

      System.out.printf("---               Property       No    Electricity           Natural Gas         \n");
      System.out.printf("---                   Type Entities     Use (kBtu) Percent    Use (kBtu) Percent \n");
      System.out.printf("--- ============================================================================ \n");

      double totalEPercentage = 0.0;
      double totalGPercentage = 0.0;

      iterator = propertyTypeDataItemMap.entrySet().iterator();
      while(iterator.hasNext()) {
         Map.Entry<String, HashSet<EnergyDataItem>> entry = iterator.next();
         String key = (String) entry.getKey();
         Set<EnergyDataItem> value = (Set) entry.getValue();

         // Compute total electicity consumption in set ...

         double totalEnergy = 0.0;
         Iterator<EnergyDataItem> ij = value.iterator(); 
         while (ij.hasNext()) {
            EnergyDataItem item = (EnergyDataItem) ij.next(); 
            String energyUsage  = item.get("Electricity Use (kBtu)");
            if (energyUsage.length() > 0)
               totalEnergy = totalEnergy + Double.parseDouble( energyUsage );
         }

         double ePercentage = 100*(totalEnergy/totalElectricityUsage);
         totalEPercentage = totalEPercentage + ePercentage; 

         // Compute total gas consumption in set ...

         double totalGas = 0.0;
         Iterator<EnergyDataItem> ik = value.iterator(); 
         while (ik.hasNext()) {
            EnergyDataItem item = (EnergyDataItem) ik.next(); 
            String energyUsage  = item.get("Natural Gas Use (kBtu)");
            if (energyUsage.length() > 0)
               totalGas = totalGas + Double.parseDouble( energyUsage );
         }

         double gPercentage = 100*(totalGas/totalGasUsage);
         totalGPercentage = totalGPercentage + gPercentage; 

         // Print summary ...

         System.out.printf("--- %22s %8d  %13.1f  %6.2f %13.1f  %6.2f \n",
                            key, value.size(), totalEnergy, ePercentage, totalGas, gPercentage );
      }

      System.out.printf("--- ============================================================================ \n");
      System.out.printf("--- Total Percentages:                             %7.2f               %7.2f \n",
                         totalEPercentage, totalGPercentage );
      System.out.printf("--- ============================================================================ \n");

      System.out.printf("\n");
      System.out.printf("=========================================================\n");
      System.out.printf("PART 07: Create lists of property-specific data items ...\n");
      System.out.printf("=========================================================\n");
      System.out.printf("\n");

      HashSet officeset     = propertyTypeDataItemMap.get( "OFFICE" );
      HashSet housingset    = propertyTypeDataItemMap.get( "MULTIFAMILY HOUSING" );
      HashSet hospitalset   = propertyTypeDataItemMap.get( "HOSPITAL" );
      HashSet universityset = propertyTypeDataItemMap.get( "COLLEGE/UNIVERSITY" );

      // Print abbreviated details of hospitals ...

      hospital   = new ArrayList<EnergyDataItem> (hospitalset);
      office     = new ArrayList<EnergyDataItem> (officeset);
      housing    = new ArrayList<EnergyDataItem> (housingset);
      university = new ArrayList<EnergyDataItem> (universityset);

      // Sort hospitals by age and Site EUI index ...
      // hospital.sort( new DemoChicagoEnergyBenchmarking.euiCompare() );

      hospital.sort( new DemoChicagoEnergyBenchmarking.ageCompare() );

      // Print sorted list of hospitals ...

      System.out.printf("\n");
      System.out.printf("List of Hospitals (Sorted by Age) ... \n");
      System.out.printf("===================================================================== \n");
      System.out.printf("ID           Community Area    Age  Floor Area Source EUI    Site EUI \n");
      System.out.printf("===================================================================== \n");

      int count = 1;
      Iterator<EnergyDataItem> ij = hospital.iterator(); 
      while (ij.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) ij.next(); 

         String id        = item.get("ID");
         String community = item.get("Community Area");
         String sourceEUI = item.get("Source EUI (kBtu/sq ft)");
         String siteEUI   = item.get("Site EUI (kBtu/sq ft)");
         String age       = item.get("Age");
         String floorarea = item.get("Gross Floor Area - Buildings (sq ft)");

         System.out.printf("%2d: %22s: %5s: %10s: %10s: %10s ...\n",
                           count, community, age, floorarea, sourceEUI, siteEUI );

         count = count + 1;
      }

      // Sort offices by EUI index ...
      // office.sort( new DemoChicagoEnergyBenchmarking.euiCompare() );

      office.sort( new DemoChicagoEnergyBenchmarking.ageCompare() );

      // Print sorted list of hospitals ...

      System.out.printf("\n");
      System.out.printf("List of Offices (Sorted by Age) ... \n");
      System.out.printf("===================================================================== \n");
      System.out.printf("ID           Community Area    Age  Floor Area Source EUI    Site EUI \n");
      System.out.printf("===================================================================== \n");

      count = 1;
      ij = office.iterator(); 
      while (ij.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) ij.next(); 

         String id        = item.get("ID");
         String community = item.get("Community Area");
         String age       = item.get("Age");
         String floorarea = item.get("Gross Floor Area - Buildings (sq ft)");
         String sourceEUI = item.get("Source EUI (kBtu/sq ft)");
         String siteEUI   = item.get("Site EUI (kBtu/sq ft)");

         System.out.printf("%2d: %22s: %5s: %10s: %10s: %10s ...\n",
                           count, community, age, floorarea, sourceEUI, siteEUI );

         count = count + 1;
      }

      // Sort multifamily housing by EUI index ...
      // housing.sort( new DemoChicagoEnergyBenchmarking.euiCompare() );

      housing.sort( new DemoChicagoEnergyBenchmarking.ageCompare() );

      // Print sorted list of multfamily housing units ...

      System.out.printf("\n");
      System.out.printf("MultiFamily Housing Units (Sorted by Age) ... \n");
      System.out.printf("===================================================================== \n");
      System.out.printf("ID           Community Area    Age  Floor Area Source EUI    Site EUI \n");
      System.out.printf("===================================================================== \n");

      count = 1;
      ij = housing.iterator(); 
      while (ij.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) ij.next(); 

         String id        = item.get("ID");
         String community = item.get("Community Area");
         String age       = item.get("Age");
         String floorarea = item.get("Gross Floor Area - Buildings (sq ft)");
         String sourceEUI = item.get("Source EUI (kBtu/sq ft)");
         String siteEUI   = item.get("Site EUI (kBtu/sq ft)");

         System.out.printf("%2d: %22s: %5s: %10s: %10s: %10s ...\n",
                           count, community, age, floorarea, sourceEUI, siteEUI );

         count = count + 1;
      }

      // Sort college/university buildings by EUI index ...
      // university.sort( new DemoChicagoEnergyBenchmarking.euiCompare() );

      university.sort( new DemoChicagoEnergyBenchmarking.ageCompare() );

      // Print sorted list of college/university buildings ...

      System.out.printf("\n");
      System.out.printf("College/University Buildings (Sorted by Age) ... \n");
      System.out.printf("===================================================================== \n");
      System.out.printf("ID           Community Area    Age  Floor Area Source EUI    Site EUI \n");
      System.out.printf("===================================================================== \n");

      count = 1;
      ij = university.iterator(); 
      while (ij.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) ij.next(); 

         String id        = item.get("ID");
         String community = item.get("Community Area");
         String age       = item.get("Age");
         String floorarea = item.get("Gross Floor Area - Buildings (sq ft)");
         String sourceEUI = item.get("Source EUI (kBtu/sq ft)");
         String siteEUI   = item.get("Site EUI (kBtu/sq ft)");

         System.out.printf("%2d: %22s: %5s: %10s: %10s: %10s ...\n",
                           count, community, age, floorarea, sourceEUI, siteEUI );

         count = count + 1;
      }

      System.out.printf("\n");
      System.out.printf("============================================================== \n");
      System.out.printf("PART 08: Assemble composite (multiple property types) list ... \n");
      System.out.printf("============================================================== \n");
      System.out.printf("\n");

      // Assemble list of office/housing/hospital/university buildings ...

      composite = new ArrayList<EnergyDataItem> ();
      composite.addAll ( housingset );
      composite.addAll ( officeset );
      composite.addAll ( hospitalset );
      composite.addAll ( universityset );

      System.out.printf("--- Composite list size = %d ... \n", composite.size() );

      System.out.printf("\n");
      System.out.printf("===================================================== \n");
      System.out.printf("PART 09: Sort elements by zip code ...    \n");
      System.out.printf("===================================================== \n");
      System.out.printf("\n");

      composite.sort( new DemoChicagoEnergyBenchmarking.zipCompare() );

      // Print details of data items ...

      System.out.printf("---   No  Zip Code        Property Type               Community \n");
      System.out.printf("--- =============================================================== \n");

      count = 1;
      Iterator iz = composite.iterator(); 
      while (iz.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) iz.next(); 
         count = count + 1;

         String zipcode   = item.get("ZIP Code");
         String proptype  = item.get("Primary Property Type");
         String community = item.get("Community Area");

         System.out.printf("   %5d  %8s %20s %23s ...\n", count, zipcode, proptype, community );
      }

      System.out.printf("--- =============================================================== \n");

      System.out.printf("\n");
      System.out.printf("===================================================== \n");
      System.out.printf("PART 10: Compute EUI zones ...    \n");
      System.out.printf("===================================================== \n");
      System.out.printf("\n");

      // Sort office/housing/hospital/university buildings by EUI index ...

      composite.sort( new DemoChicagoEnergyBenchmarking.euiCompare() );

      System.out.printf("--- No hospitals = %d ... \n", hospital.size() );
      System.out.printf("--- No offices   = %d ... \n", office.size() );
      System.out.printf("--- No multi-family housing = %d ... \n", housing.size() );
      System.out.printf("--- No university buildings = %d ... \n", university.size() );

      // Assign EUI indices to composite list ...
      // -------------------------------------------------------------------

      int noComposite = composite.size();

      count = 1;
      Iterator ik = composite.iterator(); 
      while (ik.hasNext()) {
         EnergyDataItem item = (EnergyDataItem) ik.next(); 
         int         zipcode = Integer.parseInt( item.get("ZIP Code") );	
         String    sourceEUI = item.get("Source EUI (kBtu/sq ft)");
         String    proptype  = item.get("Primary Property Type");

         if (zipcode == 60616)
            System.out.printf("%s, zip code = %6d, source EUI = %s, ", proptype, zipcode, sourceEUI );

         // Define EUI zone categories ...

         if ( count <= 0.2*noComposite ) {
            item.add ( "euiZone", "LOW" );
            if (zipcode == 60616)
               System.out.printf("Zone = LOW ...\n");
         } else if ( count > 0.2*noComposite && count <= 0.4*noComposite ) {
            item.add ( "euiZone", "ZONE2" );
            if (zipcode == 60616)
               System.out.printf("Zone = ZONE2 ...\n");
         } else if ( count > 0.4*noComposite && count <= 0.6*noComposite ) {
            item.add ( "euiZone", "ZONE3" );
            if (zipcode == 60616)
               System.out.printf("Zone = ZONE3 ...\n");
         } else if ( count > 0.6*noComposite && count <= 0.8*noComposite ) {
            item.add ( "euiZone", "ZONE4" );
            if (zipcode == 60616)
               System.out.printf("Zone = ZONE4 ...\n");
         } else {
            item.add ( "euiZone", "HIGH" );
            if (zipcode == 60616)
               System.out.printf("Zone = ZONE5 ...\n");
         }

         count = count + 1;
      }

      System.out.printf("\n");
      System.out.printf("===================================================== \n");
      System.out.printf("PART 11: Validate frequency of energy data IDs ...    \n");
      System.out.printf("===================================================== \n");
      System.out.printf("\n");

      Map map = new HashMap();
      Integer ONE = new Integer(1);

      // Create arraylist of IDs ...

      ArrayList<String> ids = new ArrayList<String>();
      for (EnergyDataItem dataitem : data) {
          ids.add ( dataitem.get("ID") );
      }

      // Build ID-frequency map ...

      Iterator iterator1 = ids.iterator();
      while ( iterator1.hasNext() != false ) {
         String key = (String) iterator1.next();
         Integer frequency = (Integer)map.get(key);
         if (frequency == null) {
            frequency = ONE;
         } else {
            int value = frequency.intValue();
            frequency = new Integer(value + 1);
         }
         map.put(key, frequency);
      }

      // Create and print an ordered treemap...

      System.out.println("TreeMap of [ word, frequency ] usage");
      System.out.println("====================================");

      Map sortedMap = new TreeMap(map);

      Iterator iterator02 = sortedMap.keySet().iterator();
      while ( iterator02.hasNext() ) {
         String    key = (String) iterator02.next();
         Integer value = (Integer) map.get( key );

         System.out.printf( "ID = %s frequency = %d \n", key, value );
      }

      System.out.println("====================================");

      System.out.printf("\n");
      System.out.printf("===================================================== \n");
      System.out.printf("PART 12: Export data in OpenStreetMap file format ... \n");
      System.out.printf("===================================================== \n");
      System.out.printf("\n");

      try {
         exportOSM01 ("./chicago-energy-data.osm");
      } catch (Exception e) {}

      System.out.printf("\n");
      System.out.printf("================================================\n");
      System.out.printf("PART 13: Export data in ARFF file format ...    \n");
      System.out.printf("================================================\n");
      System.out.printf("\n");

      // ARFF data file for energy prediction ...

      try {
         exportARFF01 ("./chicago-energy-composite01-data.arff", composite );
      }
      catch (Exception e) {}

      // ARFF data file for composite (community areas) .....

      try {
         exportARFF02 ("./chicago-energy-composite02-data.arff", composite );
      }
      catch (Exception e) {}

      // ARFF data file for composite (multifamily, hospital, university, office) .....

      try {
         exportARFF03 ("./chicago-energy-composite03-data.arff", composite );
      }
      catch (Exception e) {}

      System.out.printf("=============================================\n");
      System.out.printf("Finished !! \n");
   }

   // =========================================================
   // Export energy data in OpenStreetMap file format ...
   // =========================================================

   public static void exportOSM01 ( String osmFileName ) throws Exception {

      System.out.printf("--- Enter exportOSM01 () ...\n");
      System.out.printf("--- No nodes = %d ...\n", data.size() );

      // Check that file name has .osm extension ....

      if ( osmFileName.contains(".")) {
         String[] parts = osmFileName.split("\\.");
         String extension = (String) parts[ parts.length - 1 ];

         if (DEBUG == true) {
             System.out.printf("--- Filename extension: %s ... \n", extension );
         }

         if (extension.contains("osm") != true) {
            System.out.printf("--- ERROR: ChicagoEnergyData.exportOSM01() ... \n" );
            System.out.printf("--- ERROR: Filename extension needs to be \".osm\"! ...\n" );
         }
      } else {
         throw new IllegalArgumentException("String " + osmFileName + " does not contain dot (.)");
      }

      // Create osm file ...

      Writer fw = null;
      try {
         fw = new FileWriter( osmFileName );

         // ====================================================================
         // Write header of OSM file ...
         // ====================================================================

         String header01 = "<?xml version='1.0' encoding='UTF-8'?>";
         String header02 = "<osm version=\"0.6\" generator=\"chicagoenergydata\">";
         String bounds   = "<bounds minlat=\"41.64\" minlon=\"-88.0099999\" maxlat=\"42.09\" maxlon=\"-87.3999997\"/>";

         fw.write( header01 + "\n");
         fw.write( header02 + "\n");
         fw.write( bounds + "\n");

         // ====================================================================
         // Write file data info 
         // ====================================================================

         int count = 0;
         for (EnergyDataItem dataitem : data) {
            count = count + 1;

            // String id        = String.format("%d", count );

            String id        = dataitem.get("ID");
            String community = dataitem.get("Community Area");
            String propname  = dataitem.get("Property Name");
            String proptype  = dataitem.get("Primary Property Type");
            String age       = dataitem.get("Age");
            String floorarea = dataitem.get("Gross Floor Area - Buildings (sq ft)");
            String latitude  = dataitem.get("Latitude");
            String longitude = dataitem.get("Longitude");
            String visible   = "true";
            String version   = "1";
            String changeset = "1";
            String timestamp = "2019-01-16T15:53:28Z";
            String user      = "austin";
            String uid       = "001";

            // -------------------------------------------
            // Write header ....
            // -------------------------------------------

            fw.write( "<node " );
            fw.write( String.format("id = \"%s\" ", id ) );
            // fw.write( String.format("visible = \"%s\" ", visible ) );
            // fw.write( String.format("version = \"%s\" ", version ) );
            // fw.write( String.format("changeset = \"%s\" ", changeset ) );
            // fw.write( String.format("timestamp = \"%s\" ", timestamp ) );
            // fw.write( String.format("user = \"%s\" ", user ) );
            // fw.write( String.format("uid = \"%s\" ", uid ) );
            fw.write( String.format("lat = \"%s\" ", latitude ) );
            fw.write( String.format("lon = \"%s\"", longitude ) );
            fw.write( ">" + "\n");

            // -------------------------------------------
            // Write tags ....
            // -------------------------------------------

            if (community.length() > 0 ) {
               fw.write( String.format("   <tag k = \"community\" v = \"%s\" />", community ) );
               fw.write( "\n" );
            }

            /* ....

            if (propname.length() > 0 ) {
               fw.write( String.format("   <tag k = \"name\" v = \"%s\" />", propname ) );
               fw.write( "\n" );
            }
            .... */

            if (proptype.length() > 0 ) {
               fw.write( String.format("   <tag k = \"type\" v = \"building\" />") );
               fw.write( "\n" );
               fw.write( String.format("   <tag k = \"building\" v = \"%s\" />", proptype ) );
               fw.write( "\n" );
            }

            if (floorarea.length() > 0 ) {
               fw.write( String.format("   <tag k = \"area\" v = \"%s\" />", floorarea ) );
               fw.write( "\n" );
            }

            if (age.length() > 0 ) {
               fw.write( String.format("   <tag k = \"age\" v = \"%s\" />", age ) );
               fw.write( "\n" );
            }

            // -------------------------------------------
            // Write tail ....
            // -------------------------------------------

            fw.write( "</node>" + "\n" );
         }

         // ====================================================================
         // Write tail of OSM file ....
         // ====================================================================

         String tail01   = "</osm>";
         fw.write( tail01 + "\n");

      }
      catch (IOException e) {}
      finally {
         try { fw.close(); }
         catch (Exception e) {}
      }

      System.out.printf("--- Leave exportOSM01 () ...\n");
   }

   // ===================================================================
   // ExportARFF01(): Generate ARFF file format for EUI Zone analysis ...
   // ===================================================================

   public static void exportARFF01 ( String arffFileName, ArrayList<EnergyDataItem> data ) throws Exception {

      System.out.printf("--- Enter exportARFF01 () ...\n");

      // Check that file name has .arff extension ....

      if ( arffFileName.contains(".")) {
         String[] parts = arffFileName.split("\\.");
         String extension = (String) parts[ parts.length - 1 ];
         if (DEBUG == true) {
             System.out.printf("--- Filename extension: %s ... \n", extension );
         }

         if (extension.contains("arff") != true) {
            System.out.printf("--- ERROR: ChicagoEnergyData.exportARFF01() ... \n" );
            System.out.printf("--- ERROR: Filename extension needs to be \".arff\"! ...\n" );
         }
      } else {
         throw new IllegalArgumentException("String " + arffFileName + " does not contain dot (.)");
      }

      // Create arff file ...

      Writer fw = null;
      try {
         fw = new FileWriter( arffFileName );

         // Write file header info 

         String header01 = "% ====================================== ... \n";
         String header02 = "% Energy Data File         January, 2019 ... \n";
         String header03 = "% ====================================== ... \n";

         fw.write( header01 );
         fw.write( header02 );
         fw.write( header03 );

         // ====================================================================
         // Write data file relation 
         // ====================================================================

         // @relation energy

         String relation = String.format("@relation chicagoenergy");

         fw.write( relation );
         fw.write( "\n");

         // ====================================================================
         // Write data file attributes 
         // ====================================================================

         String communityAreaList = "'ALBANY PARK', 'ARCHER HEIGHTS', 'ARMOUR SQUARE', ASHBURN, 'AUBURN GRESHAM', AUSTIN, 'AVALON PARK', AVONDALE, 'BELMONT CRAGIN', BEVERLY, BRIDGEPORT, 'BRIGHTON PARK', BURNSIDE, 'CALUMET HEIGHTS', CHATHAM, 'CHICAGO LAWN', CLEARING, DOUGLAS, DUNNING, 'EAST GARFIELD PARK', 'EAST SIDE', EDGEWATER, 'EDISON PARK', ENGLEWOOD, 'FOREST GLEN', 'FULLER PARK', 'GAGE PARK', 'GARFIELD RIDGE', 'GRAND BOULEVARD', 'GREATER GRAND CROSSING', HEGEWISCH, HERMOSA, 'HUMBOLDT PARK', 'HYDE PARK', 'IRVING PARK', 'JEFFERSON PARK', KENWOOD, 'LAKE VIEW', 'LINCOLN PARK', 'LINCOLN SQUARE', 'LOGAN SQUARE', LOOP, 'LOWER WEST SIDE', 'MCKINLEY PARK', MONTCLARE, 'MORGAN PARK', 'MOUNT GREENWOOD', 'NEAR NORTH SIDE', 'NEAR SOUTH SIDE', 'NEAR WEST SIDE', 'NEW CITY', 'NORTH CENTER', 'NORTH LAWNDALE', 'NORTH PARK', 'NORWOOD PARK', OAKLAND, OHARE, 'PORTAGE PARK', PULLMAN, RIVERDALE, 'ROGERS PARK', ROSELAND, 'SOUTH CHICAGO', 'SOUTH DEERING', 'SOUTH LAWNDALE', 'SOUTH SHORE', UPTOWN, 'WASHINGTON HEIGHTS', 'WASHINGTON PARK', 'WEST ELSDON', 'WEST ENGLEWOOD', 'WEST GARFIELD PARK', 'WEST LAWN', 'WEST PULLMAN', 'WEST RIDGE', 'WEST TOWN', WOODLAWN";

    //   String propertyTypeList = "'ADULT EDUCATION', 'AUTO DEALERSHIP', 'BANK BRANCH', COLLEGE/UNIVERSITY, 'CONVENTION CENTER', COURTHOUSE, EDUCATION, 'ENCLOSED MALL', 'FINANCIAL OFFICE', 'FITNESS CENTER', HOSPITAL, HOTEL, 'INDOOR ARENA', 'K-12 SCHOOL', LABORATORY, LIBRARY, 'LIFESTYLE CENTER', MALL, 'MEDICAL OFFICE', 'MIXED USE PROPERTY', 'MOVIE THEATER', 'MULTIFAMILY HOUSING', MUSEUM, OFFICE, OTHER, 'PERFORMING ARTS', 'PHYSICAL THERAPY', 'PRE-SCHOOL/DAYCARE', 'PRISON', 'PUBLIC ASSEMBLY', 'PUBLIC SERVICES', RECREATION, 'REPAIR SERVICES', 'RESIDENCE HALL', 'RESIDENTIAL', 'RESIDENTIAL CARE', 'RETAIL STORE', 'SENIOR CARE', SERVICES, 'SOCIAL/MEETING HALL', 'SPECIALTY HOSPITAL', 'STRIP MALL', 'SUPERMARKET/GROCERY', 'SURGICAL CENTER', 'URGENT CARE CLINIC', 'WHOLESALE CLUB', 'WORSHIP FACILITY'"; 

         String propertyTypeList = "'COLLEGE/UNIVERSITY', 'HOSPITAL', 'MULTIFAMILY HOUSING', 'OFFICE'"; 

         // Assemble string for list of zip codes ... 

         HashSet<String> zipCodeItems = new HashSet<String>();

         Iterator<EnergyDataItem> ii = data.iterator(); 
         while (ii.hasNext()) {
            EnergyDataItem dataitem = (EnergyDataItem) ii.next(); 
            String zipitem = String.format( "'%s'", dataitem.get("ZIP Code"));
            zipCodeItems.add( zipitem );
         }

         TreeSet<String> zipCodesTree = new TreeSet<String>( zipCodeItems );
         Iterator<String> ij = zipCodesTree.iterator(); 

         String zipCodesList = "";
         boolean first = true;
         while (ij.hasNext()) {
            String zipitem = (String) ij.next(); 
            if (first == true ) {
               zipCodesList = zipCodesList + zipitem;
               first = false;
            } else {
               zipCodesList = zipCodesList + "," + zipitem;
            }
         }

         // Define attribute strings ...

         String attribute01 = String.format("@attribute community {%s}", communityAreaList );
         String attribute02 = String.format("@attribute propertytype {%s}", propertyTypeList );
         String attribute03 = String.format("@attribute zip {%s}", zipCodesList );
         String attribute04 = String.format("@attribute age numeric");
         String attribute05 = String.format("@attribute area numeric");
         String attribute06 = String.format("@attribute latitude numeric");
         String attribute07 = String.format("@attribute longitude numeric");
         String attribute08 = String.format("@attribute class {LOW, ZONE2, ZONE3, ZONE4, HIGH}");
         String attribute09 = String.format("@attribute eui numeric");

         // community area ...

         // fw.write( attribute01 );
         // fw.write( "\n");

         // property type ...

         // fw.write( attribute02 );
         // fw.write( "\n");

         // latitude ...

         // fw.write( attribute06 );
         // fw.write( "\n");

         // longitude ...

         // fw.write( attribute07 );
         // fw.write( "\n");

         // zip code ...

         fw.write( attribute03 );
         fw.write( "\n");

         // age ...

         fw.write( attribute04 );
         fw.write( "\n");

         // floor area ...

         fw.write( attribute05 );
         fw.write( "\n");

         // eui numeric ...
         // fw.write( attribute09 );
         // fw.write( "\n");

         // eui zone ...

         fw.write( attribute08 );
         fw.write( "\n");

         // Write file data info 

         String marker = String.format("@data");
         fw.write( marker );
         fw.write( "\n");

         System.out.printf("--- No data items = %d ... \n", data.size() );

         Iterator<EnergyDataItem> ik = data.iterator(); 
         while (ik.hasNext() == true) {
            EnergyDataItem dataitem = (EnergyDataItem) ik.next(); 

            String community = String.format("'%s'", dataitem.get("Community Area"));
            String proptype  = String.format("'%s'", dataitem.get("Primary Property Type"));
            String zip       = String.format( "%6s", dataitem.get("ZIP Code"));
            String age       = String.format( "%5s", dataitem.get("Age"));
            String floorarea = String.format("%12s", dataitem.get("Gross Floor Area - Buildings (sq ft)"));
            String latitude  = String.format("%12s", dataitem.get("Latitude"));
            String longitude = String.format("%13s", dataitem.get("Longitude"));
            String euizone   = String.format( "%7s", dataitem.get("euiZone"));
            String sourceeui = String.format( "%7s", dataitem.get("Source EUI (kBtu/sq ft)"));

            // fw.write( proptype.toUpperCase() );
            // fw.write( community.toUpperCase() );
            // fw.write( "," );
            // fw.write( "," );
            // fw.write( floorarea );
            // fw.write( "," );
            // fw.write( latitude );
            // fw.write( longitude );

            fw.write( zip );
            fw.write( "," );
            fw.write( age );
            fw.write( "," );
            fw.write( floorarea );
            fw.write( "," );
            fw.write( euizone );
            fw.write( "\n");

            // fw.write( sourceeui );
            // fw.write( "," );

         }

      }
      catch (IOException e) {}
      finally {
         try { fw.close(); }
         catch (Exception e) {}
      }

      System.out.printf("--- Leave exportARFF01 () ...\n");
   }

   // =========================================================
   // ExportARFF02(): Generate composite ARFF file format ...
   // =========================================================

   public static void exportARFF02 ( String arffFileName, ArrayList<EnergyDataItem> data ) throws Exception {

      System.out.printf("--- Enter exportARFF02 () ...\n");

      // Check that file name has .arff extension ....

      if ( arffFileName.contains(".")) {
         String[] parts = arffFileName.split("\\.");
         String extension = (String) parts[ parts.length - 1 ];
         if (DEBUG == true) {
             System.out.printf("--- Filename extension: %s ... \n", extension );
         }

         if (extension.contains("arff") != true) {
            System.out.printf("--- ERROR: ChicagoEnergyData.exportARFF01() ... \n" );
            System.out.printf("--- ERROR: Filename extension needs to be \".arff\"! ...\n" );
         }
      } else {
         throw new IllegalArgumentException("String " + arffFileName + " does not contain dot (.)");
      }

      // Create arff file ...

      Writer fw = null;
      try {
         fw = new FileWriter( arffFileName );

         // Write file header info 

         String header01 = "% ============================================================= ... \n";
         String header02 = "% Composite property type data file               January, 2019 ... \n";
         String header03 = "% ============================================================= ... \n";
         String header04 = "% \n";

         fw.write( header01 );
         fw.write( header02 );
         fw.write( header03 );
         fw.write( header04 );

         // ====================================================================
         // Write data file relation 
         // ====================================================================

         // @relation energy

         String relation = String.format("@relation neighborhoods");

         fw.write( relation );
         fw.write( "\n");

         // ====================================================================
         // Write data file attributes 
         // ====================================================================

         String communityAreaList = "'ALBANY PARK', 'ARCHER HEIGHTS', 'ARMOUR SQUARE', ASHBURN, 'AUBURN GRESHAM', AUSTIN, 'AVALON PARK', AVONDALE, 'BELMONT CRAGIN', BEVERLY, BRIDGEPORT, 'BRIGHTON PARK', BURNSIDE, 'CALUMET HEIGHTS', CHATHAM, 'CHICAGO LAWN', CLEARING, DOUGLAS, DUNNING, 'EAST GARFIELD PARK', 'EAST SIDE', EDGEWATER, 'EDISON PARK', ENGLEWOOD, 'FOREST GLEN', 'FULLER PARK', 'GAGE PARK', 'GARFIELD RIDGE', 'GRAND BOULEVARD', 'GREATER GRAND CROSSING', HEGEWISCH, HERMOSA, 'HUMBOLDT PARK', 'HYDE PARK', 'IRVING PARK', 'JEFFERSON PARK', KENWOOD, 'LAKE VIEW', 'LINCOLN PARK', 'LINCOLN SQUARE', 'LOGAN SQUARE', LOOP, 'LOWER WEST SIDE', 'MCKINLEY PARK', MONTCLARE, 'MORGAN PARK', 'MOUNT GREENWOOD', 'NEAR NORTH SIDE', 'NEAR SOUTH SIDE', 'NEAR WEST SIDE', 'NEW CITY', 'NORTH CENTER', 'NORTH LAWNDALE', 'NORTH PARK', 'NORWOOD PARK', OAKLAND, OHARE, 'PORTAGE PARK', PULLMAN, RIVERDALE, 'ROGERS PARK', ROSELAND, 'SOUTH CHICAGO', 'SOUTH DEERING', 'SOUTH LAWNDALE', 'SOUTH SHORE', UPTOWN, 'WASHINGTON HEIGHTS', 'WASHINGTON PARK', 'WEST ELSDON', 'WEST ENGLEWOOD', 'WEST GARFIELD PARK', 'WEST LAWN', 'WEST PULLMAN', 'WEST RIDGE', 'WEST TOWN', WOODLAWN";

         String propertyTypeList = "COLLEGE/UNIVERSITY, HOSPITAL, 'MULTIFAMILY HOUSING', OFFICE"; 

         // Assemble string for list of zip codes ... 

         HashSet<String> zipCodeItems = new HashSet<String>();

         Iterator<EnergyDataItem> ii = data.iterator(); 
         while (ii.hasNext()) {
            EnergyDataItem dataitem = (EnergyDataItem) ii.next(); 
            String zipitem = String.format( "'%s'", dataitem.get("ZIP Code"));
            zipCodeItems.add( zipitem );
         }

         TreeSet<String> zipCodesTree = new TreeSet<String>( zipCodeItems );
         Iterator<String> ij = zipCodesTree.iterator(); 

         String zipCodesList = "";
         boolean first = true;
         while (ij.hasNext()) {
            String zipitem = (String) ij.next(); 
            if (first == true ) {
               zipCodesList = zipCodesList + zipitem;
               first = false;
            } else {
               zipCodesList = zipCodesList + "," + zipitem;
            }
         }

         // Define attribute strings ...

         String attribute01 = String.format("@attribute community {%s}", communityAreaList );
         String attribute02 = String.format("@attribute propertytype {%s}", propertyTypeList );
         String attribute03 = String.format("@attribute zip {%s}", zipCodesList );
         String attribute04 = String.format("@attribute age numeric");
         String attribute05 = String.format("@attribute area numeric");
         String attribute06 = String.format("@attribute latitude numeric");
         String attribute07 = String.format("@attribute longitude numeric");
         String attribute08 = String.format("@attribute eui numeric");
         String attribute09 = String.format("@attribute class {LOW, MEDIUM, HIGH}");

         fw.write( attribute02 );
         fw.write( "\n");
         fw.write( attribute03 );
         fw.write( "\n");
         fw.write( attribute04 );
         fw.write( "\n");
         fw.write( attribute05 );
         fw.write( "\n");
         fw.write( attribute08 );
         fw.write( "\n");
         fw.write( attribute01 );
         fw.write( "\n");

         // fw.write( attribute06 );
         // fw.write( "\n");
         // fw.write( attribute07 );
         // fw.write( "\n");
         // fw.write( attribute09 );
         // fw.write( "\n");

         // Write file data info 

         String marker = String.format("@data");
         fw.write( marker );
         fw.write( "\n");

         System.out.printf("--- No data items = %d ... \n", data.size() );

         Iterator<EnergyDataItem> ik = data.iterator(); 
         while (ik.hasNext() == true) {
            EnergyDataItem dataitem = (EnergyDataItem) ik.next(); 

            String community = String.format("'%s'", dataitem.get("Community Area"));
            String proptype  = String.format("'%s'", dataitem.get("Primary Property Type"));
            String zip       = String.format( "%6s", dataitem.get("ZIP Code"));
            String age       = String.format( "%5s", dataitem.get("Age"));
            String floorarea = String.format("%12s", dataitem.get("Gross Floor Area - Buildings (sq ft)"));

            // String floor01 = String.format("%12s", dataitem.get("Gross Floor Area - Buildings (sq ft)"));
            // double farea   = Math.log( Double.parseDouble( floor01 ));
            // String floorarea = String.format( "%s", farea );

            String latitude  = String.format("%12s", dataitem.get("Latitude"));
            String longitude = String.format("%13s", dataitem.get("Longitude"));
            String euizone   = String.format( "%7s", dataitem.get("euiZone"));
            String sourceeui = String.format( "%7s", dataitem.get("Source EUI (kBtu/sq ft)"));

            fw.write( proptype.toUpperCase() );
            fw.write( "," );
            fw.write( zip );
            fw.write( "," );
            fw.write( age );
            fw.write( "," );
            fw.write( floorarea );
            fw.write( "," );
            fw.write( sourceeui );
            fw.write( "," );
            fw.write( community.toUpperCase() );
            fw.write( "\n");
         }

      }
      catch (IOException e) {}
      finally {
         try { fw.close(); }
         catch (Exception e) {}
      }

      System.out.printf("--- Leave exportARFF02 () ...\n");
   }

   // =========================================================
   // ExportARFF03(): Generate composite ARFF file format ...
   // =========================================================

   public static void exportARFF03 ( String arffFileName, ArrayList<EnergyDataItem> data ) throws Exception {

      System.out.printf("--- Enter exportARFF03 () ...\n");

      // Check that file name has .arff extension ....

      if ( arffFileName.contains(".")) {
         String[] parts = arffFileName.split("\\.");
         String extension = (String) parts[ parts.length - 1 ];
         if (DEBUG == true) {
             System.out.printf("--- Filename extension: %s ... \n", extension );
         }

         if (extension.contains("arff") != true) {
            System.out.printf("--- ERROR: ChicagoEnergyData.exportARFF01() ... \n" );
            System.out.printf("--- ERROR: Filename extension needs to be \".arff\"! ...\n" );
         }
      } else {
         throw new IllegalArgumentException("String " + arffFileName + " does not contain dot (.)");
      }

      // Create arff file ...

      Writer fw = null;
      try {
         fw = new FileWriter( arffFileName );

         // Write file header info 

         String header01 = "% ============================================================= ... \n";
         String header02 = "% Composite property type data file               January, 2019 ... \n";
         String header03 = "% ============================================================= ... \n";
         String header04 = "% \n";

         fw.write( header01 );
         fw.write( header02 );
         fw.write( header03 );
         fw.write( header04 );

         // ====================================================================
         // Write data file relation 
         // ====================================================================

         // @relation energy

         String relation = String.format("@relation neighborhoods");

         fw.write( relation );
         fw.write( "\n");

         // ====================================================================
         // Write data file attributes 
         // ====================================================================

         String communityAreaList = "'ALBANY PARK', 'ARCHER HEIGHTS', 'ARMOUR SQUARE', ASHBURN, 'AUBURN GRESHAM', AUSTIN, 'AVALON PARK', AVONDALE, 'BELMONT CRAGIN', BEVERLY, BRIDGEPORT, 'BRIGHTON PARK', BURNSIDE, 'CALUMET HEIGHTS', CHATHAM, 'CHICAGO LAWN', CLEARING, DOUGLAS, DUNNING, 'EAST GARFIELD PARK', 'EAST SIDE', EDGEWATER, 'EDISON PARK', ENGLEWOOD, 'FOREST GLEN', 'FULLER PARK', 'GAGE PARK', 'GARFIELD RIDGE', 'GRAND BOULEVARD', 'GREATER GRAND CROSSING', HEGEWISCH, HERMOSA, 'HUMBOLDT PARK', 'HYDE PARK', 'IRVING PARK', 'JEFFERSON PARK', KENWOOD, 'LAKE VIEW', 'LINCOLN PARK', 'LINCOLN SQUARE', 'LOGAN SQUARE', LOOP, 'LOWER WEST SIDE', 'MCKINLEY PARK', MONTCLARE, 'MORGAN PARK', 'MOUNT GREENWOOD', 'NEAR NORTH SIDE', 'NEAR SOUTH SIDE', 'NEAR WEST SIDE', 'NEW CITY', 'NORTH CENTER', 'NORTH LAWNDALE', 'NORTH PARK', 'NORWOOD PARK', OAKLAND, OHARE, 'PORTAGE PARK', PULLMAN, RIVERDALE, 'ROGERS PARK', ROSELAND, 'SOUTH CHICAGO', 'SOUTH DEERING', 'SOUTH LAWNDALE', 'SOUTH SHORE', UPTOWN, 'WASHINGTON HEIGHTS', 'WASHINGTON PARK', 'WEST ELSDON', 'WEST ENGLEWOOD', 'WEST GARFIELD PARK', 'WEST LAWN', 'WEST PULLMAN', 'WEST RIDGE', 'WEST TOWN', WOODLAWN";

         String propertyTypeList = "COLLEGE/UNIVERSITY, HOSPITAL, 'MULTIFAMILY HOUSING', OFFICE"; 

         // Assemble string for list of zip codes ... 

         HashSet<String> zipCodeItems = new HashSet<String>();

         Iterator<EnergyDataItem> ii = data.iterator(); 
         while (ii.hasNext()) {
            EnergyDataItem dataitem = (EnergyDataItem) ii.next(); 
            String zipitem = String.format( "'%s'", dataitem.get("ZIP Code"));
            zipCodeItems.add( zipitem );
         }

         TreeSet<String> zipCodesTree = new TreeSet<String>( zipCodeItems );
         Iterator<String> ij = zipCodesTree.iterator(); 

         String zipCodesList = "";
         boolean first = true;
         while (ij.hasNext()) {
            String zipitem = (String) ij.next(); 
            if (first == true ) {
               zipCodesList = zipCodesList + zipitem;
               first = false;
            } else {
               zipCodesList = zipCodesList + "," + zipitem;
            }
         }

         // Define attribute strings ...

         String attribute01 = String.format("@attribute community {%s}", communityAreaList );
         String attribute02 = String.format("@attribute propertytype {%s}", propertyTypeList );
         String attribute03 = String.format("@attribute zip {%s}", zipCodesList );
         String attribute04 = String.format("@attribute age numeric");
         String attribute05 = String.format("@attribute area numeric");
         String attribute06 = String.format("@attribute latitude numeric");
         String attribute07 = String.format("@attribute longitude numeric");
         String attribute08 = String.format("@attribute eui numeric");
         String attribute09 = String.format("@attribute class {LOW, MEDIUM, HIGH}");

         fw.write( attribute01 );
         fw.write( "\n");
         fw.write( attribute03 );
         fw.write( "\n");
         fw.write( attribute04 );
         fw.write( "\n");
         fw.write( attribute05 );
         fw.write( "\n");
         fw.write( attribute08 );
         fw.write( "\n");
         fw.write( attribute02 );
         fw.write( "\n");

         // fw.write( attribute06 );
         // fw.write( "\n");
         // fw.write( attribute07 );
         // fw.write( "\n");
         // fw.write( attribute09 );
         // fw.write( "\n");

         // Write file data info 

         String marker = String.format("@data");
         fw.write( marker );
         fw.write( "\n");

         System.out.printf("--- No data items = %d ... \n", data.size() );

         Iterator<EnergyDataItem> ik = data.iterator(); 
         while (ik.hasNext() == true) {
            EnergyDataItem dataitem = (EnergyDataItem) ik.next(); 

            String community = String.format("'%s'", dataitem.get("Community Area"));
            String proptype  = String.format("'%s'", dataitem.get("Primary Property Type"));
            String zip       = String.format( "%6s", dataitem.get("ZIP Code"));
            String age       = String.format( "%5s", dataitem.get("Age"));
            String floorarea = String.format("%12s", dataitem.get("Gross Floor Area - Buildings (sq ft)"));

            // String floor01 = String.format("%12s", dataitem.get("Gross Floor Area - Buildings (sq ft)"));
            // double farea   = Math.log( Double.parseDouble( floor01 ));
            // String floorarea = String.format( "%s", farea );

            String latitude  = String.format("%12s", dataitem.get("Latitude"));
            String longitude = String.format("%13s", dataitem.get("Longitude"));
            String euizone   = String.format( "%7s", dataitem.get("euiZone"));
            String sourceeui = String.format( "%7s", dataitem.get("Source EUI (kBtu/sq ft)"));

            fw.write( community.toUpperCase() );
            fw.write( "," );
            fw.write( zip );
            fw.write( "," );
            fw.write( age );
            fw.write( "," );
            fw.write( floorarea );
            fw.write( "," );
            fw.write( sourceeui );
            fw.write( "," );
            fw.write( proptype.toUpperCase() );
            fw.write( "\n");

            // fw.write( latitude );
            // fw.write( "," );
            // fw.write( longitude );
            // fw.write( "," );
            // fw.write( euizone );
            // fw.write( "\n");

         }

      }
      catch (IOException e) {}
      finally {
         try { fw.close(); }
         catch (Exception e) {}
      }

      System.out.printf("--- Leave exportARFF03 () ...\n");
   }
}
