/*
 *  =======================================================================
 *  Polyline.java: Ordered set of connected line segments.
 *  =======================================================================
 */

package geometry;

import java.lang.Math;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;

public class Polyline {
   protected String sName;
   private BoundingBox box;              // bounding box for line segment ...
   private WayPoint source      = null;  // source way point ...
   private WayPoint destination = null;  // destination way point ...
   private double pathwayLength;
   protected boolean isClosed;
   boolean DEBUG = true;

   // ArrayList of way points ...

   private ArrayList<WayPoint> pathway = null;

   // Constructor methods ...

   public Polyline() {
      pathway       = new ArrayList<WayPoint>();
      box           = new BoundingBox();
      pathwayLength = 0.0;
      this.isClosed = false;
   }

   // --------------------------------------------
   // Add way point to polyline model ...
   // --------------------------------------------

   public void add ( WayPoint node ) {
      pathway.add( node );

      // Set source and destination nodes ...

      if ( pathway.size() == 1 ) {
         source = destination = pathway.get(0);
      } else  
         destination = pathway.get( pathway.size() - 1 );

      // Update bounding box ...

      box.addPoint ( node.getX(), node.getY() );

      // Update pathway length ...

      if ( pathway.size() > 1 ) {
         pathwayLength = getPathwayLength();
      }
   }

   // --------------------------------------------
   // Set/get pathway name ...
   // --------------------------------------------

   public String getName() {
      return sName;
   }

   public void setName( String sName ) {
      this.sName = sName;
   }

   // Get pathway as an array list model ...
	
   public ArrayList getPathway() {
      return pathway;
   }

   // -------------------------------------------------------------------
   // Compute pathway length ...
   // -------------------------------------------------------------------

   public double getPathwayLength() {
      double length = 0.0;

      System.out.printf("--- Enter getPathwayLength() ...\n");
      System.out.printf("--- pathway.size() = %d ...\n", pathway.size() );

      for(int i = 0; i < pathway.size() - 1; i++) {

         double x1 = pathway.get(i).getX();
         double y1 = pathway.get(i).getY();
         double x2 = pathway.get(i+1).getX();
         double y2 = pathway.get(i+1).getY();

         double distance = Math.sqrt(Math.pow(x2-x1,2) + Math.pow(y2-y1,2));

         System.out.printf("--- Edge: %2d: length = %6.2f ...\n", i+1, distance );

         length = length + distance;
      }

      return length;
   }
	
   // -------------------------------------------------------------------
   // Process waypoints long the pathway ...
   // -------------------------------------------------------------------

   public void processPathway() {
      double length = getPathwayLength();

      System.out.printf("--- \n");
      System.out.printf("--- Enter processPathway() ...\n");
      System.out.printf("--- ---------------------- ...\n");

      for(int i = 0; i < pathway.size() - 1; i++) {

         System.out.printf("--- Edge: %d ...\n", (i+1) );

         double x1 = pathway.get(i).getX();
         double y1 = pathway.get(i).getY();
         double x2 = pathway.get(i+1).getX();
         double y2 = pathway.get(i+1).getY();

         double distance = Math.sqrt(Math.pow(x2-x1,2) + Math.pow(y2-y1,2));
         System.out.printf("--- Length = %8.2f ...\n", distance );

         // Set "distance from source" and "distance from destination" parameters ...

         WayPoint wptStart = pathway.get(i);
         WayPoint wptEnd   = pathway.get(i+1);

         double startDistance = wptStart.getDistanceFromSource();
         double endDistance = startDistance + distance;

         wptStart.setDistanceFromSource ( startDistance );
         wptEnd.setDistanceFromSource ( endDistance );

         wptStart.setDistanceFromDestination ( length - startDistance );
         wptEnd.setDistanceFromDestination ( length - endDistance );
      }

      System.out.printf("--- ---------------------- ...\n");
      System.out.printf("--- Leave processPathway() ...\n");
      System.out.printf("--- \n");
   }
	
   // Retrieve source and destination way points ...

   public WayPoint getSource() {
      return source;
   }

   public WayPoint getDestination() {
      return destination;
   }

   // Retrieve length of pathway ....

   public double getLength() {
      return pathwayLength;
   }

   // Compute length of segment (x0,y0) -- (x1,y1)...

   public double segmentLength( double dx0, double dy0, double dx1, double dy1 ) {
       return Math.sqrt( (dx1 - dx0)*(dx1 - dx0) + (dy1 - dy0)*(dy1 - dy0) );
   }

   // Compute angle for coordinates in four quadrants ....

   public double getAngle( double dX, double dY ) {
       double angle = 0.0;

       if (DEBUG == true) {
          System.out.printf("In getAngle(): dX = %f dY = %f ...\n", dX, dY );
       }

       if ( dY >= 0.0 && dX >= 0.0 )
            angle = Math.atan( dY/dX );
       if ( dY >= 0.0 && dX < 0.0 )
            angle = Math.PI + Math.atan( dY/dX );
       if ( dY  < 0.0 && dX < 0.0 )
            angle = Math.PI + Math.atan( dY/dX );
       if ( dY  < 0.0 && dX >= 0.0 )
            angle = 2*Math.PI + Math.atan( dY/dX );

       if (DEBUG == true) {
          System.out.printf("In getAngle(): angle = %f ...\n", angle );
       }

       return angle;
   }

   // ===================================================================
   // Return node point that is a given distance along pathway.
   // ===================================================================

   public Node2D pointAtDistance( double length ) {
      boolean foundSegment = false;
      WayPoint start  = pathway.get(0);
      WayPoint finish = pathway.get(0);

      if ( DEBUG == true ) {
         System.out.println("--- Entering Pathway.pointAtDistance() ...");
         System.out.printf( "--- length = %f ... \n", length );
      }

      // Check that the length value is valid ...

      if ( length < 0.0 || length > pathwayLength ) {
         System.out.printf("--- ERROR in Pathway.pointAtDistance()");
         System.out.printf("--- Invalid Input: length = %f ... \n", length );
      }

      // Walk along waypoint list and find segment within which the point lies ...

      int j = 1;
      while ( j < pathway.size() && foundSegment == false) {
         start  = pathway.get(j-1);
         finish = pathway.get(j);

         if ( DEBUG == true ) {
            System.out.printf( "--- j = %d ... \n", j );
            System.out.printf( "--- start distance from origin  = %f ... \n",   start.getDistanceFromSource() );
            System.out.printf( "--- finish distance from origin = %f ... \n", finish.getDistanceFromSource() );
         }

         if ( length >= start.getDistanceFromSource() && length  < finish.getDistanceFromSource()) {
              foundSegment = true;
         }
         j = j + 1;
      }

      // Compute coordinates at offset distance ....

      double offset = length - start.getDistanceFromSource();
      double theta  = getAngle(finish.getX() - start.getX(),
                               finish.getY() - start.getY() );

      double xPoint = (double) (start.getX() + offset * Math.cos(theta));
      double yPoint = (double) (start.getY() + offset * Math.sin(theta));

      Node2D c = new Node2D( xPoint, yPoint );

      if ( DEBUG == true ) {
         System.out.printf( "*** Returning : %s ... \n", c.toString() );
      }
      
      return c;
   }

   // ===================================================================
   // Return angle that is a given distance along pathway.
   // ===================================================================

   public double angleAtDistance( double length ) {
      boolean foundSegment = false;
      WayPoint start  = pathway.get(0);
      WayPoint finish = pathway.get(0);

      // Check that the length value is valid ...

      if ( length < 0.0 || length > pathwayLength ) {
        System.out.printf("--- ERROR in Polyline.pointAtDistance()");
        System.out.printf("--- Invalid Input: length = %f ... \n", length );
      }

      // Walk along waypoint list and find segment within which the point lies ...

      int j = 1;
      while ( j < pathway.size() && foundSegment == false) {
         start  = pathway.get(j-1);
         finish = pathway.get(j);
         if ( length >= start.getDistanceFromSource() &&
              length < finish.getDistanceFromSource() ) {
              foundSegment = true;
              if (DEBUG == true )
                 System.out.printf("*** Found segment: j = %d ...\n", j );
         }
         j = j + 1;
      }

      // Compute offset distance ....

      double offset = length - start.getDistanceFromSource();

      // Compute the slope.

      double xstart  = start.getX();
      double ystart  = start.getY();
      double xfinish = finish.getX();
      double yfinish = finish.getY();

      if (DEBUG == true ) {
         System.out.printf(" xstart  = %f ystart  = %f ...\n",  xstart,  ystart );
         System.out.printf(" xfinish = %f yfinish = %f ...\n", xfinish, yfinish );
      }

      double dX = xfinish - xstart;
      double dY = yfinish - ystart;
      double theta = getAngle( dX, dY );

      if (DEBUG == true ) {
         System.out.printf(" Returning: theta  = %f ...\n", theta );
      }

      return theta;
   }

   // ======================================================================
   // Create string description of Polyline ...
   // ======================================================================

   public String toString() {
      StringBuffer buf = new StringBuffer();

      buf.append("\n");
      buf.append("--- geometry.Polyline() ... \n");
      buf.append("--- ============================================ \n"); 

      if ( this.sName == null ) 
         buf.append("--- Name = null ... \n");
      else
         buf.append("--- Name = " + this.sName + "\n");

      if (this.isClosed == true )
         buf.append("--- Is closed?   = true ... \n"); 
      else
         buf.append("--- Is closed?   = false ... \n"); 

      buf.append( String.format("--- Source         = %s  \n", source.getName() ) );
      buf.append( String.format("--- Destination    = %s  \n", destination.getName() ) );
      buf.append( String.format("--- Pathway length = %f  \n", pathwayLength ) );
      buf.append( String.format("--- %s ... \n", box.toString() ) );
      buf.append("--- ============================================ \n"); 

      for(int i = 0; i < pathway.size(); i = i + 1 ) {
          String wpt = pathway.get(i).toString();
          buf.append( String.format("%s \n", wpt ) );
      }

      buf.append("--- ============================================ \n"); 

      return buf.toString();
   }
}
